@@ -1,5 +1,5 @@
 /* Locale-specific memory comparison.
-   Copyright (C) 1999, 2002, 2003 Free Software Foundation, Inc.
+   Copyright (C) 1999, 2002, 2003, 2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,24 +13,18 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 /* Contributed by Paul Eggert <eggert@twinsun.com>.  */
 
-#if HAVE_CONFIG_H
+#ifdef HAVE_CONFIG_H
 # include <config.h>
 #endif
 
 #include "memcoll.h"
 
 #include <errno.h>
-#ifndef errno
-extern int errno;
-#endif
-
-#if HAVE_STRING_H
-# include <string.h>
-#endif
+#include <string.h>
 
 /* Compare S1 (with length S1LEN) and S2 (with length S2LEN) according
    to the LC_COLLATE locale.  S1 and S2 do not overlap, and are not
@@ -44,40 +38,49 @@ memcoll (char *s1, size_t s1len, char *s2, size_t s2len)
 
 #if HAVE_STRCOLL
 
-  char n1 = s1[s1len];
-  char n2 = s2[s2len];
+  /* strcoll is slow on many platforms, so check for the common case
+     where the arguments are bytewise equal.  Otherwise, walk through
+     the buffers using strcoll on each substring.  */
 
-  s1[s1len++] = '\0';
-  s2[s2len++] = '\0';
-
-  while (! (errno = 0, (diff = strcoll (s1, s2)) || errno))
+  if (s1len == s2len && memcmp (s1, s2, s1len) == 0)
+    diff = 0;
+  else
     {
-      /* strcoll found no difference, but perhaps it was fooled by NUL
-	 characters in the data.  Work around this problem by advancing
-	 past the NUL chars.  */
-      size_t size1 = strlen (s1) + 1;
-      size_t size2 = strlen (s2) + 1;
-      s1 += size1;
-      s2 += size2;
-      s1len -= size1;
-      s2len -= size2;
+      char n1 = s1[s1len];
+      char n2 = s2[s2len];
 
-      if (s1len == 0)
+      s1[s1len++] = '\0';
+      s2[s2len++] = '\0';
+
+      while (! (errno = 0, (diff = strcoll (s1, s2)) || errno))
 	{
-	  if (s2len != 0)
-	    diff = -1;
-	  break;
-	}
-      else if (s2len == 0)
-	{
-	  diff = 1;
-	  break;
+	  /* strcoll found no difference, but perhaps it was fooled by NUL
+	     characters in the data.  Work around this problem by advancing
+	     past the NUL chars.  */
+	  size_t size1 = strlen (s1) + 1;
+	  size_t size2 = strlen (s2) + 1;
+	  s1 += size1;
+	  s2 += size2;
+	  s1len -= size1;
+	  s2len -= size2;
+
+	  if (s1len == 0)
+	    {
+	      if (s2len != 0)
+		diff = -1;
+	      break;
+	    }
+	  else if (s2len == 0)
+	    {
+	      diff = 1;
+	      break;
+	    }
 	}
+
+      s1[s1len - 1] = n1;
+      s2[s2len - 1] = n2;
     }
 
-  s1[s1len - 1] = n1;
-  s2[s2len - 1] = n2;
-
 #else
 
   diff = memcmp (s1, s2, s1len < s2len ? s1len : s2len);
