@@ -4,10 +4,9 @@ static char const G_line[] =
   "# The keywords COLOR, OPTIONS, and EIGHTBIT (honored by the\0"
   "# slackware version of dircolors) are recognized but ignored.\0"
   "# Below, there should be one TERM entry for each termtype that is colorizable\0"
-  "TERM linux\0"
-  "TERM linux-c\0"
-  "TERM mach-color\0"
-  "TERM console\0"
+  "TERM Eterm\0"
+  "TERM ansi\0"
+  "TERM color-xterm\0"
   "TERM con132x25\0"
   "TERM con132x30\0"
   "TERM con132x43\0"
@@ -18,18 +17,30 @@ static char const G_line[] =
   "TERM con80x43\0"
   "TERM con80x50\0"
   "TERM con80x60\0"
+  "TERM console\0"
   "TERM cygwin\0"
   "TERM dtterm\0"
+  "TERM gnome\0"
+  "TERM konsole\0"
+  "TERM kterm\0"
+  "TERM linux\0"
+  "TERM linux-c\0"
+  "TERM mach-color\0"
+  "TERM mlterm\0"
   "TERM putty\0"
-  "TERM xterm\0"
-  "TERM xterm-color\0"
-  "TERM xterm-debian\0"
   "TERM rxvt\0"
+  "TERM rxvt-cygwin\0"
+  "TERM rxvt-cygwin-native\0"
+  "TERM rxvt-unicode\0"
   "TERM screen\0"
   "TERM screen-bce\0"
   "TERM screen-w\0"
+  "TERM screen.linux\0"
   "TERM vt100\0"
-  "TERM Eterm\0"
+  "TERM xterm\0"
+  "TERM xterm-256color\0"
+  "TERM xterm-color\0"
+  "TERM xterm-debian\0"
   "# Below are the color init strings for the basic file types. A color init\0"
   "# string consists of one or more of the following numeric codes:\0"
   "# Attribute codes:\0"
@@ -48,7 +59,7 @@ static char const G_line[] =
   "DOOR 01;35 # door\0"
   "BLK 40;33;01 # block device driver\0"
   "CHR 40;33;01 # character device driver\0"
-  "ORPHAN 40;31;01 # symlink to nonexistent file\0"
+  "ORPHAN 40;31;01 # symlink to nonexistent file, or non-stat'able file\0"
   "SETUID 37;41 # file that is setuid (u+s)\0"
   "SETGID 30;43 # file that is setgid (g+s)\0"
   "STICKY_OTHER_WRITABLE 30;42 # dir that is sticky and other-writable (+t,o+w)\0"
