@@ -1,5 +1,5 @@
 /* install - copy files and set attributes
-   Copyright (C) 89, 90, 91, 1995-2004 Free Software Foundation, Inc.
+   Copyright (C) 89, 90, 91, 1995-2006 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 /* Written by David MacKenzie <djm@gnu.ai.mit.edu> */
 
@@ -31,10 +31,12 @@
 #include "cp-hash.h"
 #include "copy.h"
 #include "dirname.h"
-#include "makepath.h"
+#include "filenamecat.h"
+#include "mkancesdirs.h"
+#include "mkdir-p.h"
 #include "modechange.h"
-#include "path-concat.h"
 #include "quote.h"
+#include "stat-time.h"
 #include "utimens.h"
 #include "xstrtol.h"
 
@@ -47,14 +49,6 @@
 # include <sys/wait.h>
 #endif
 
-struct passwd *getpwnam ();
-struct group *getgrnam ();
-
-#ifndef _POSIX_VERSION
-uid_t getuid ();
-gid_t getgid ();
-#endif
-
 #if ! HAVE_ENDGRENT
 # define endgrent() ((void) 0)
 #endif
@@ -72,18 +66,20 @@ gid_t getgid ();
 /* Number of bytes of a file to copy at a time. */
 #define READ_SIZE (32 * 1024)
 
-static bool change_timestamps (const char *from, const char *to);
-static bool change_attributes (const char *path);
+static bool change_timestamps (struct stat const *from_sb, char const *to);
+static bool change_attributes (char const *name);
 static bool copy_file (const char *from, const char *to,
 		       const struct cp_options *x);
-static bool install_file_to_path (const char *from, const char *to,
-				  const struct cp_options *x);
+static bool install_file_in_file_parents (char const *from, char *to,
+					  struct cp_options *x);
 static bool install_file_in_dir (const char *from, const char *to_dir,
 				 const struct cp_options *x);
 static bool install_file_in_file (const char *from, const char *to,
 				  const struct cp_options *x);
 static void get_ids (void);
-static void strip (const char *path);
+static void strip (char const *name);
+static void announce_mkdir (char const *dir, void *options);
+static int make_ancestor (char const *dir, void *options);
 void usage (int status);
 
 /* The name this program was run with, for error messages. */
@@ -103,9 +99,20 @@ static char *group_name;
 /* The group ID corresponding to `group_name'. */
 static gid_t group_id;
 
-/* The permissions to which the files will be set.  The umask has
+#define DEFAULT_MODE (S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)
+
+/* The file mode bits to which non-directory files will be set.  The umask has
    no effect. */
-static mode_t mode = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
+static mode_t mode = DEFAULT_MODE;
+
+/* Similar, but for directories.  */
+static mode_t dir_mode = DEFAULT_MODE;
+
+/* The file mode bits that the user cares about.  This should be a
+   superset of DIR_MODE and a subset of CHMOD_MODE_BITS.  This matters
+   for directories, since otherwise directories may keep their S_ISUID
+   or S_ISGID bits.  */
+static mode_t dir_mode_bits = CHMOD_MODE_BITS;
 
 /* If true, strip executable files after copying them. */
 static bool strip_files;
@@ -125,7 +132,6 @@ static struct option const long_options[] =
   {"strip", no_argument, NULL, 's'},
   {"suffix", required_argument, NULL, 'S'},
   {"target-directory", required_argument, NULL, 't'},
-  {"version-control", required_argument, NULL, 'V'}, /* Deprecated. FIXME. */
   {"verbose", no_argument, NULL, 'v'},
   {GETOPT_HELP_OPTION_DECL},
   {GETOPT_VERSION_OPTION_DECL},
@@ -142,7 +148,7 @@ cp_option_init (struct cp_options *x)
   x->hard_link = false;
   x->interactive = I_UNSPECIFIED;
   x->move_mode = false;
-  x->myeuid = geteuid ();
+  x->chown_privileges = chown_privileges ();
   x->one_file_system = false;
   x->preserve_ownership = false;
   x->preserve_links = false;
@@ -161,7 +167,6 @@ cp_option_init (struct cp_options *x)
   x->mode = S_IRUSR | S_IWUSR;
   x->stdin_tty = false;
 
-  x->umask_kill = 0;
   x->update = false;
   x->verbose = false;
   x->dest_info = NULL;
@@ -176,7 +181,7 @@ cp_option_init (struct cp_options *x)
 static bool
 target_directory_operand (char const *file)
 {
-  char const *b = base_name (file);
+  char const *b = last_component (file);
   size_t blen = strlen (b);
   bool looks_like_a_dir = (blen == 0 || ISSLASH (b[blen - 1]));
   struct stat st;
@@ -225,18 +230,11 @@ main (int argc, char **argv)
      we'll actually use backup_suffix_string.  */
   backup_suffix_string = getenv ("SIMPLE_BACKUP_SUFFIX");
 
-  while ((optc = getopt_long (argc, argv, "bcsDdg:m:o:pt:TvV:S:", long_options,
+  while ((optc = getopt_long (argc, argv, "bcsDdg:m:o:pt:TvS:", long_options,
 			      NULL)) != -1)
     {
       switch (optc)
 	{
-	case 'V':  /* FIXME: this is deprecated.  Remove it in 2001.  */
-	  error (0, 0,
-		 _("warning: --version-control (-V) is obsolete;  support for\
- it\nwill be removed in some future release.  Use --backup=%s instead."
-		   ), optarg);
-	  /* Fall through.  */
-
 	case 'b':
 	  make_backups = true;
 	  if (optarg)
@@ -353,12 +351,12 @@ main (int argc, char **argv)
 
   if (specified_mode)
     {
-      struct mode_change *change = mode_compile (specified_mode, 0);
-      if (change == MODE_INVALID)
+      struct mode_change *change = mode_compile (specified_mode);
+      if (!change)
 	error (EXIT_FAILURE, 0, _("invalid mode %s"), quote (specified_mode));
-      else if (change == MODE_MEMORY_EXHAUSTED)
-	xalloc_die ();
-      mode = mode_adjust (0, change);
+      mode = mode_adjust (0, false, 0, change, NULL);
+      dir_mode = mode_adjust (0, true, 0, change, &dir_mode_bits);
+      free (change);
     }
 
   get_ids ();
@@ -367,11 +365,9 @@ main (int argc, char **argv)
     {
       int i;
       for (i = 0; i < n_files; i++)
-	{
-	  ok &=
-	    make_path (file[i], mode, mode, owner_id, group_id, false,
-		       (x.verbose ? _("creating directory %s") : NULL));
-	}
+	ok &= make_dir_parents (file[i], make_ancestor, &x,
+				dir_mode, announce_mkdir,
+				dir_mode_bits, owner_id, group_id, false);
     }
   else
     {
@@ -382,7 +378,7 @@ main (int argc, char **argv)
       if (!target_directory)
         {
           if (mkdir_and_install)
-	    ok = install_file_to_path (file[0], file[1], &x);
+	    ok = install_file_in_file_parents (file[0], file[1], &x);
 	  else
 	    ok = install_file_in_file (file[0], file[1], &x);
 	}
@@ -404,30 +400,16 @@ main (int argc, char **argv)
    Return true if successful.  */
 
 static bool
-install_file_to_path (const char *from, const char *to,
-		      const struct cp_options *x)
+install_file_in_file_parents (char const *from, char *to,
+			      struct cp_options *x)
 {
-  char *dest_dir = dir_name (to);
-  bool ok = true;
-
-  /* Make sure that the parent of the destination is a directory.  */
-  if (! STREQ (dest_dir, "."))
+  if (mkancesdirs (to, make_ancestor, x) != 0)
     {
-      /* Someone will probably ask for a new option or three to specify
-	 owner, group, and permissions for parent directories.  Remember
-	 that this option is intended mainly to help installers when the
-	 distribution doesn't provide proper install rules.  */
-#define DIR_MODE (S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)
-      ok = make_path (dest_dir, DIR_MODE, DIR_MODE, owner_id, group_id, true,
-		      (x->verbose ? _("creating directory %s") : NULL));
+      error (0, errno, _("cannot create directory %s"), to);
+      return false;
     }
 
-  free (dest_dir);
-
-  if (ok)
-    ok = install_file_in_file (from, to, x);
-
-  return ok;
+  return install_file_in_file (from, to, x);
 }
 
 /* Copy file FROM onto file TO and give TO the appropriate
@@ -438,14 +420,20 @@ static bool
 install_file_in_file (const char *from, const char *to,
 		      const struct cp_options *x)
 {
+  struct stat from_sb;
+  if (x->preserve_timestamps && stat (from, &from_sb) != 0)
+    {
+      error (0, errno, _("cannot stat %s"), quote (from));
+      return false;
+    }
   if (! copy_file (from, to, x))
     return false;
   if (strip_files)
     strip (to);
   if (! change_attributes (to))
     return false;
-  if (x->preserve_timestamps)
-    return change_timestamps (from, to);
+  if (x->preserve_timestamps && (strip_files || ! S_ISREG (from_sb.st_mode)))
+    return change_timestamps (&from_sb, to);
   return true;
 }
 
@@ -457,8 +445,8 @@ static bool
 install_file_in_dir (const char *from, const char *to_dir,
 		     const struct cp_options *x)
 {
-  const char *from_base = base_name (from);
-  char *to = path_concat (to_dir, from_base, NULL);
+  const char *from_base = last_component (from);
+  char *to = file_name_concat (to_dir, from_base, NULL);
   bool ret = install_file_in_file (from, to, x);
   free (to);
   return ret;
@@ -481,14 +469,12 @@ copy_file (const char *from, const char *to, const struct cp_options *x)
   return copy (from, to, false, x, &copy_into_self, NULL);
 }
 
-/* Set the attributes of file or directory PATH.
+/* Set the attributes of file or directory NAME.
    Return true if successful.  */
 
 static bool
-change_attributes (const char *path)
+change_attributes (char const *name)
 {
-  bool ok = true;
-
   /* chown must precede chmod because on some systems,
      chown clears the set[ug]id bits for non-superusers,
      resulting in incorrect permissions.
@@ -499,47 +485,29 @@ change_attributes (const char *path)
      the install command is that the file is supposed to end up with
      precisely the attributes that the user specified (or defaulted).
      If the file doesn't end up with the group they asked for, they'll
-     want to know.  But AFS returns EPERM when you try to change a
-     file's group; thus the kludge.  */
+     want to know.  */
 
-  if (chown (path, owner_id, group_id)
-#ifdef AFS
-      && errno != EPERM
-#endif
-      )
-    {
-      error (0, errno, _("cannot change ownership of %s"), quote (path));
-      ok = false;
-    }
+  if (! (owner_id == (uid_t) -1 && group_id == (gid_t) -1)
+      && chown (name, owner_id, group_id) != 0)
+    error (0, errno, _("cannot change ownership of %s"), quote (name));
+  else if (chmod (name, mode) != 0)
+    error (0, errno, _("cannot change permissions of %s"), quote (name));
+  else
+    return true;
 
-  if (ok && chmod (path, mode))
-    {
-      error (0, errno, _("cannot change permissions of %s"), quote (path));
-      ok = false;
-    }
-
-  return ok;
+  return false;
 }
 
 /* Set the timestamps of file TO to match those of file FROM.
    Return true if successful.  */
 
 static bool
-change_timestamps (const char *from, const char *to)
+change_timestamps (struct stat const *from_sb, char const *to)
 {
-  struct stat stb;
   struct timespec timespec[2];
+  timespec[0] = get_stat_atime (from_sb);
+  timespec[1] = get_stat_mtime (from_sb);
 
-  if (stat (from, &stb))
-    {
-      error (0, errno, _("cannot obtain time stamps for %s"), quote (from));
-      return false;
-    }
-
-  timespec[0].tv_sec = stb.st_atime;
-  timespec[0].tv_nsec = TIMESPEC_NS (stb.st_atim);
-  timespec[1].tv_sec = stb.st_mtime;
-  timespec[1].tv_nsec = TIMESPEC_NS (stb.st_mtim);
   if (utimens (to, timespec))
     {
       error (0, errno, _("cannot set time stamps for %s"), quote (to));
@@ -548,14 +516,14 @@ change_timestamps (const char *from, const char *to)
   return true;
 }
 
-/* Strip the symbol table from the file PATH.
+/* Strip the symbol table from the file NAME.
    We could dig the magic number out of the file first to
    determine whether to strip it, but the header files and
    magic numbers vary so much from system to system that making
    it portable would be very difficult.  Not worth the effort. */
 
 static void
-strip (const char *path)
+strip (char const *name)
 {
   int status;
   pid_t pid = fork ();
@@ -566,7 +534,7 @@ strip (const char *path)
       error (EXIT_FAILURE, errno, _("fork system call failed"));
       break;
     case 0:			/* Child. */
-      execlp ("strip", "strip", path, NULL);
+      execlp ("strip", "strip", name, NULL);
       error (EXIT_FAILURE, errno, _("cannot run strip"));
       break;
     default:			/* Parent. */
@@ -624,6 +592,25 @@ get_ids (void)
     group_id = (gid_t) -1;
 }
 
+/* Report that directory DIR was made, if OPTIONS requests this.  */
+static void
+announce_mkdir (char const *dir, void *options)
+{
+  struct cp_options const *x = options;
+  if (x->verbose)
+    error (0, 0, _("creating directory %s"), quote (dir));
+}
+
+/* Make ancestor directory DIR, with options OPTIONS.  */
+static int
+make_ancestor (char const *dir, void *options)
+{
+  int r = mkdir (dir, DEFAULT_MODE);
+  if (r == 0)
+    announce_mkdir (dir, options);
+  return r;
+}
+
 void
 usage (int status)
 {
@@ -649,7 +636,7 @@ In the 4th form, create all components of the given DIRECTORY(ies).\n\
 Mandatory arguments to long options are mandatory for short options too.\n\
 "), stdout);
       fputs (_("\
-      --backup[=CONTROL] make a backup of each existing destination file\n\
+      --backup[=CONTROL]  make a backup of each existing destination file\n\
   -b                  like --backup but does not accept an argument\n\
   -c                  (ignored)\n\
   -d, --directory     treat all arguments as directory names; create all\n\
